/* -*- mode: c++ -*-

  This file is part of the Life library

  Author(s): Christophe Prud'homme <christophe.prudhomme@ujf-grenoble.fr>
       Date: 2005-11-27

  Copyright (C) 2005,2006 EPFL

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 3.0 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
// $Id: petsc_linear_solver.h,v 1.2 2005/02/22 22:17:34 jwpeterson Exp $

// The libMesh Finite Element Library.
// Copyright (C) 2002-2005  Benjamin S. Kirk, John W. Peterson

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 3.0 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
/**
   \file solverlinearpetsc.hpp
   \author Christophe Prud'homme <christophe.prudhomme@ujf-grenoble.fr>
   \date 2005-11-27
 */

#ifndef __petsc_linear_solver_h__
#define __petsc_linear_solver_h__


#include <life/lifecore/life.hpp>


#include <life/lifealg/solverlinear.hpp>
#include <life/lifealg/matrixpetsc.hpp>
#include <life/lifealg/vectorpetsc.hpp>


/**
 * Petsc include files.
 */
#ifdef HAVE_PETSC_H

#ifndef USE_COMPLEX_NUMBERS
extern "C" {
# include <petscversion.h>
# if (PETSC_VERSION_MAJOR == 2) && (PETSC_VERSION_MINOR <= 1)
#   include <petscsles.h>
# else
#   include <petscksp.h>
# endif
}
#else
# include <petscversion.h>
# if (PETSC_VERSION_MAJOR == 2) && (PETSC_VERSION_MINOR <= 1)
#   include <petscsles.h>
# else
#   include <petscksp.h>
# endif
#endif

namespace Life
{


/**
 * This class provides an interface to PETSc
 * iterative solvers that is compatible with the \p libMesh
 * \p LinearSolver<>
 *
 * @author Benjamin Kirk, 2002-2005
 * @author Christophe Prud'homme, 2005
 */

template <typename T>
class SolverLinearPetsc : public SolverLinear<T>
{
    typedef SolverLinear<T> super;

public:

    typedef typename super::value_type value_type;
    typedef typename super::real_type real_type;

    /**
     *  Constructor. Initializes Petsc data structures
     */
    SolverLinearPetsc ();

    /**
     * Destructor.
     */
    ~SolverLinearPetsc ();

    /**
     * Release all memory and clear data structures.
     */
    void clear ();

    /**
     * Initialize data structures if not done so already.
     */
    void init ();

    /**
     * Call the Petsc solver.  It calls the method below, using the
     * same matrix for the system and preconditioner matrices.
     *
     * \param mat System Matrix
     * \param prec Preconditioning Matrix
     * \param x Solution vector
     * \param b RHS vector
     * \param tolerance Stopping tolerance
     * \param maxit maximum Number of Iterations
     * \param transpose true to solve the transpose system, false otherwise
     */
    std::pair<unsigned int, real_type>
    solve (MatrixSparse<T>  const &mat,
           Vector<T> & x,
           Vector<T> const& b,
           const double tolerance,
           const unsigned int maxit,
           bool transpose )
    {
        return this->solve( mat, mat, x, b, tolerance, maxit, transpose );
    }

    /**
     * This method allows you to call a linear solver while specifying
     * the matrix to use as the (left) preconditioning matrix.  Note
     * that the linear solver will not compute a preconditioner in
     * this case, and will instead premultiply by the matrix you
     * provide.
     *
     * In PETSc, this is accomplished by calling
     *
     * PCSetType(_pc, PCMAT);
     *
     * before invoking KSPSolve().  Note: this functionality is not
     * implemented in the SolverLinear class since there is not a
     * built-in analog to this method for LasPack -- You could
     * probably implement it by hand if you wanted.
     *
     *
     * \param mat System Matrix
     * \param prec Preconditioning Matrix
     * \param x Solution vector
     * \param b RHS vector
     * \param tolerance Stopping tolerance
     * \param maxit maximum Number of Iterations
     * \param transpose true to solve the transpose system, false otherwise
     */
    std::pair<unsigned int, real_type>
    solve (MatrixSparse<T>  const& mat,
           MatrixSparse<T>  const& prec,
           Vector<T> & x,
           Vector<T> const& b,
           const double tolerance,
           const unsigned int maxit,
           bool transpose );

    /**
     * Returns the raw PETSc preconditioner context pointer.  This allows
     * you to specify the PCShellSetApply() and PCShellSetSetUp() functions
     * if you desire.  Just don't do anything crazy like calling PCDestroy()!
     */
    PC pc() { this->init(); return _M_pc; }

    /**
     * Fills the input vector with the sequence of residual norms
     * from the latest iterative solve.
     */
    void getResidualHistory(std::vector<double>& hist);

    /**
     * Returns just the initial residual for the solve just
     * completed with this interface.  Use this method instead
     * of the one above if you just want the starting residual
     * and not the entire history.
     */
    real_type getInitialResidual();

private:

    mpi::communicator M_comm;

    /**
     * Tells PETSC to use the user-specified solver stored in
     * \p _solver_type
     */
    void setPetscSolverType ();

    /**
     * Tells PETSC to use the user-specified preconditioner stored in
     * \p _preconditioner_type
     */
    void setPetscPreconditionerType ();

    // SLES removed from >= PETSc 2.2.0
#if (PETSC_VERSION_MAJOR == 2) && (PETSC_VERSION_MINOR <= 1)

    /**
     * Linear solver context
     */
    SLES _M_sles;

#endif

    /**
     * Preconditioner context
     */
    PC _M_pc;

    /**
     * Krylov subspace context
     */
    KSP _M_ksp;
};


/*----------------------- functions ----------------------------------*/
template <typename T>
inline
SolverLinearPetsc<T>::SolverLinearPetsc ()
{
  if (M_comm.size() == 1)
    this->setPreconditionerType(  LU_PRECOND );
  else
    this->setPreconditionerType( BLOCK_JACOBI_PRECOND );
}



template <typename T>
inline
SolverLinearPetsc<T>::~SolverLinearPetsc ()
{
  this->clear ();
}
} // Life


#endif // #ifdef HAVE_PETSC_H
#endif // #ifdef __petsc_linear_solver_h__

